@@ -237,12 +237,12 @@ static FILE *Oops_objdump(const char *file, const OPTIONS *options)
 /* Forward references. */
 static const char *Oops_arch_to_eip(const OPTIONS *options);
 static void Oops_decode_one_add(SYMBOL_SET *ss, const addr_t address,
-				const char *line1, const OPTIONS *options);
+				const char *line1, const char type, const OPTIONS *options);
 static addr_t Oops_truncate_address(addr_t address, const OPTIONS *options);
 
 /* Process one code line from objdump, ignore everything else */
 static void Oops_decode_one(SYMBOL_SET *ss, const char *line, addr_t eip,
-			    int adjust, const OPTIONS *options)
+			    int adjust, const char type, const OPTIONS *options)
 {
     int i;
     addr_t address, eip_relative;
@@ -314,11 +314,11 @@ static void Oops_decode_one(SYMBOL_SET *ss, const char *line, addr_t eip,
 	    malloc_error(procname);
 	snprintf(line2, i, "%s %s %s",
 	    line, format_address(eip_relative, options), map);
-	Oops_decode_one_add(ss, address, line2, options);
+	Oops_decode_one_add(ss, address, line2, type, options);
 	free(line2);
     }
     else
-	Oops_decode_one_add(ss, address, line, options);  /* as is */
+	Oops_decode_one_add(ss, address, line, type, options);  /* as is */
     re_strings_free(&re_Oops_objdump, &string);
 }
 
@@ -453,7 +453,7 @@ static int Oops_code_values(const unsigned char* code_text, unsigned char *code,
  * happy, replace XXX with an target specific string.
  */
 static void Oops_decode_one_add(SYMBOL_SET *ss, const addr_t address,
-				const char *line1, const OPTIONS *options)
+				const char *line1, const char type, const OPTIONS *options)
 {
     int i;
     const char *p1;
@@ -484,7 +484,7 @@ static void Oops_decode_one_add(SYMBOL_SET *ss, const addr_t address,
     if (strlen(line2) > strlen(line1))
 	ERROR("line2 overrun, should never happen line1='%s'\nline2='%s'",
 	    line1, line2);
-    add_symbol_n(ss, address, 'C', 1, line2);
+    add_symbol_n(ss, address, type, 1, line2);
     free(line2);
 }
 
@@ -638,10 +638,6 @@ static void Oops_set_regs(const char *line, const char *p,
 static regex_t     re_Oops_cris_regs;
 static regmatch_t *re_Oops_cris_regs_pmatch;
 static int         s390_reg_num = -1;
-#define MIPS64_CKSEG0_MASK    	0xffffffff80000000ULL
-#define MIPS64_CKSEG0_BASE    	0xffffffff80000000ULL
-#define MIPS64_KSEG0_BASE    	0x0000000080000000ULL
-#define MIPS64_XKPHYS5_BASE    	0xa800000000000000ULL
 
 /* All address arithmetic is done in 64 bit mode, truncate to the result to
  * a valid range for the target.
@@ -658,20 +654,6 @@ static addr_t Oops_truncate_address(addr_t address, const OPTIONS *options)
 	    u32 &= 0x7fffffff;  /* Really 31 bit addressing on i370/s390 */
 	address = u32;
     }
-    if (options->target &&
-	strstr(options->target, "mips") &&
-	strstr(options->target, "64")) {
-	/*
-	 * Some address gymnastics are necessary due to the weird way
-	 * in which mips64 kernel are built as elf32 then converted
-	 * to elf64 after linking.
-	 */
-	if ((address & MIPS64_CKSEG0_MASK) == MIPS64_CKSEG0_BASE ||
-	    (address & MIPS64_CKSEG0_MASK) == MIPS64_KSEG0_BASE) {
-	    /* normalise physical addrs in 32-bit kernel segments */
-	    address = (address & ~MIPS64_CKSEG0_MASK) | MIPS64_XKPHYS5_BASE;
-	}
-    }
     address &= truncate_mask;
     return(address);
 }
@@ -1765,6 +1747,7 @@ static int Oops_print(const char *line, const char **text, char ***string,
 	    "|<[0-9]+>"			/* kmsg */
 	    "|[0-9]+\\|[^|]+\\|[^|]+\\|[^|]+\\|"	/* syslog-ng */
 	    "|[0-9][AC] [0-9]{3} [0-9]{3}[cir][0-9]{2}:"	/* SNIA */
+	    "\\+"			/* leading '+' */
 	    ")",
 	REG_NEWLINE|REG_EXTENDED|REG_ICASE,
 	&re_Oops_prefix_pmatch);
@@ -1925,10 +1908,14 @@ static int Oops_print(const char *line, const char **text, char ***string,
     /* ppc */	    "|GPR[0-9]+: "
     /* any other ppc lines to print? */
 
+    /* mips */	    "|.* in .*\\.[ch]:.*, line [0-9]+:"
     /* mips */	    "|\\$[0-9 ]+:"
-    /* mips */	    "|epc "
-    /* mips */	    "|Status:"
-    /* mips */	    "|Cause :"
+    /* mips */	    "|Hi *:"
+    /* mips */	    "|Lo *:"
+    /* mips */	    "|epc *:"
+    /* mips */	    "|badvaddr *:"
+    /* mips */	    "|Status *:"
+    /* mips */	    "|Cause *:"
     /* any other mips lines to print? */
 
     /* arm */	    "|Backtrace:"
@@ -1994,7 +1981,7 @@ static int Oops_print(const char *line, const char **text, char ***string,
     /* ia64 */	    "|MCA"
     /* ia64 */	    "|SAL"
     /* ia64 */	    "|Processor State"
-    /* ia64 */	    "|Bank [0-9]+ "
+    /* ia64 */	    "|Bank [0-9]+"
     /* ia64 */	    "|Register Stack"
     /* ia64 */	    "|Processor Error Info"
     /* ia64 */	    "|proc err map"
@@ -2208,22 +2195,21 @@ static const char *Oops_code(const char *line, char ***string, int string_max)
 /*                      End architecture sensitive code                       */
 /******************************************************************************/
 
-/* Decode the Oops Code: via objdump. */
-static void Oops_decode(const unsigned char* code_text, addr_t eip,
-			SYMBOL_SET *ss, char ***string, int string_max,
-			const bfd *ibfd, OPTIONS *options)
+/* Decode part or all of the Oops Code: via objdump. */
+static void Oops_decode_part(const char* code, int code_size,
+			     addr_t eip, int adjust,
+			     const char type, SYMBOL_SET *ss,
+			     char ***string, int string_max,
+			     const bfd *ibfd, OPTIONS *options)
 {
     FILE *f;
-    char *file, *line = NULL, code[CODE_SIZE];
-    int size = 0, adjust, lines = 0;
-    static char const procname[] = "Oops_decode";
+    char *file, *line = NULL;
+    int size = 0, lines = 0;
+    static char const procname[] = "Oops_decode_part";
 
     DEBUG(1, "%s", "");
-    /* text to binary */
-    if (!Oops_code_values(code_text, code, &adjust, options))
-	return;
     /* binary to same format as ksymoops */
-    if (!(file = Oops_code_to_file(code, CODE_SIZE, ibfd, options)))
+    if (!(file = Oops_code_to_file(code, code_size, ibfd, options)))
 	return;
     /* objdump the pseudo object */
     if (!(f = Oops_objdump(file, options)))
@@ -2231,7 +2217,7 @@ static void Oops_decode(const unsigned char* code_text, addr_t eip,
     while (fgets_local(&line, &size, f, procname)) {
 	DEBUG(2, "%s", line);
 	++lines;
-	Oops_decode_one(ss, line, eip, adjust, options);
+	Oops_decode_one(ss, line, eip, adjust, type, options);
     }
     pclose_local(f, procname);  /* opened in Oops_objdump */
     free(line);
@@ -2243,6 +2229,28 @@ static void Oops_decode(const unsigned char* code_text, addr_t eip,
     }
 }
 
+/* Decode the Oops Code: via objdump. */
+static void Oops_decode(const unsigned char* code_text, addr_t eip,
+			SYMBOL_SET *ss, char ***string, int string_max,
+			const bfd *ibfd, OPTIONS *options)
+{
+    char code[CODE_SIZE];
+    int adjust;
+    static char const procname[] = "Oops_decode";
+
+    DEBUG(1, "%s", "");
+    /* text to binary */
+    if (!Oops_code_values(code_text, code, &adjust, options))
+	return;
+    /* adjust is -ve number of bytes before eip */
+    if (adjust && options->vli) {
+	    Oops_decode_part(code, -adjust, eip, adjust, 'U', ss, string, string_max, ibfd, options);
+	    Oops_decode_part(code-adjust, CODE_SIZE+adjust, eip, 0, 'C', ss, string, string_max, ibfd, options);
+    } else {
+	    Oops_decode_part(code, CODE_SIZE, eip, adjust, 'C', ss, string, string_max, ibfd, options);
+    }
+}
+
 /* Reached the end of an Oops report, format the extracted data. */
 static void Oops_format(const SYMBOL_SET *ss_format, OPTIONS *options)
 {
@@ -2257,15 +2265,21 @@ static void Oops_format(const SYMBOL_SET *ss_format, OPTIONS *options)
     compare_Version();  /* Oops might have a version one day */
     printf("\n");
     for (s = ss_format->symbol, i = 0; i < ss_format->used; ++i, ++s) {
-	/* For type C data, print Code:, address, map, "name" (actually
+	/* For type C and U data, print Code:, address, map, "name" (actually
 	 * the text of an objdump line).  For type I (id) print the string.
 	 * For other types print name, address, map.
 	 */
-	if (s->type == 'C') {
-	    if (prev_type != s->type)
+	if (s->type == 'C' || s->type == 'U') {
+	    if (prev_type != s->type) {
 		printf("\n");
+		if (s->type == 'U')
+		    printf("This architecture has variable length instructions, decoding before eip\n"
+			   "is unreliable, take these instructions with a pinch of salt.\n\n");
+		else if (prev_type == 'U')
+		    printf("This decode from eip onwards should be reliable\n\n");
+	    }
 	    eye_catcher = "";
-	    if (eip && i < ss_format->used-1 && (s+1)->type == 'C' &&
+	    if (eip && i < ss_format->used-1 && (s+1)->type == s->type &&
 		eip >= s->address && eip < (s+1)->address)
 		eye_catcher = "   <=====";
 	    if (options->short_lines)
@@ -2398,6 +2412,7 @@ int Oops_read(OPTIONS *options)
 		    if ((ret_addr = Oops_alpha_ra(text, &string, MAX_STRINGS)))
 			Oops_set_alpha_ra(ret_addr, &ss_format);
 		    Oops_set_eip(start, &eip, &ss_format, me, ibfd, options);
+		    options->vli = strstr(line, " VLI") != NULL;
 		}
 		if (Oops_eip_alpha_spinlock(text, &string, MAX_STRINGS,
 					    options)) {
